/***************************************
    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    common.h -- miscellaneous definitions.
					
***************************************/


#ifndef COMMON_H
#define COMMON_H		/*+ To stop multiple inclusions. + */

#include "../config.h"

#include <stdio.h>
#include <sys/types.h>

#if STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#elif HAVE_STRINGS_H
# include <strings.h>
#endif				/* STDC_HEADERS */

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif				/* HAVE_SYS_TIME_H */
#endif				/* TIME_WITH_SYS_TIME */

#include "str.h"

/* Defines */
#ifdef DEBUG
#define debug(S) fprintf(stderr,S)
#define debug2(S,T) fprintf(stderr,S,T)
#else
#define debug(S)
#define debug2(S,T)
#endif

#define debug_err(S)  fprintf(stderr, "ERROR: \"%s\" at line %d of %s\n", S, \
                      __LINE__, __FILE__);
#define debug_perr(S) fprintf(stderr, \
                      "ERROR: \"%s: %s\" at line %d of %s\n", S, \
                      errno<sys_nerr ? sys_errlist[errno] : "Unknown error", \
                      __LINE__, __FILE__);

#define TRUE 1
#define FALSE 0

#define DM    242
#define IP    244
#define WILL  251
#define WONT  252
#define DO    253
#define DONT  254
#define IAC   255

#define BUF_LEN 4097
#define MAX_LINE_LEN 180	/* For control connections */

#define GET_CLNT 1
#define GET_SRVR 2
#define GET_BOTH 3

/*Log levels */
#define INFO   20		/* Who connects, etc. */
#define IMPORT 15		/* Startup messages as well. */
#define ERROR  10		/* Non critical errors */
#define ATTACK  5		/* Suspicious stuff */

/* Error codes */
typedef enum _exitreason {
	TIMEOUT,
	CLIENT_CLOSE,
	SERVER_CLOSE,
	NO_CLIENT_DATA,		/* Unable to connect to it */
	NO_SERVER_DATA,
	CLIENT_RUBBISH,		/* Malformed control stream */
	SERVER_RUBBISH,
	MALLOC,
	SOCKET
} exitreason;


typedef enum _socketuse {
	ACTV,
	PASV,
	CTRL
} socketuse;

typedef struct _connection {
	struct sockaddr_in address;
	int fd;

	sstr *buf;
} connection;

typedef struct _session_info {
	connection client_control;
	connection server_control;
	connection client_data;
	connection server_data;

	enum { ACTIVE, PASSIVE, APCONV } mode;
	enum { UPLOAD, DOWNLOAD, NEITHER } state;
	int listen;		/*fd we are listening on */
	int anon;
} session_info;

typedef struct {
	pid_t pid;
	enum { SNAT, DNAT, XSNAT, XDNAT, FAILED, SUCCEEDED, FLUSH } type;
	struct sockaddr_in src;
	struct sockaddr_in dest;
	struct sockaddr_in to;
} mangle_req;

int sufd;			/* if ==-1 --> We aren't doing transparent data proxying 
				   == 0 --> Kernel 2.2 - no need for it
				   > 0 --> A socket for sending mangle_reqs over */

#include "configs.h"
/*****************
**    misc.c    **
******************/
int listen_on_socket(struct sockaddr_in *listen_address, socketuse use);
int connect_to_socket(struct sockaddr_in listen_address, socketuse use);
int bind_me(int fd, struct sockaddr_in *addr, int lo, int hi,
	    int attempts);
struct sockaddr_in com2n(int a1, int a2, int a3, int a4, int p1, int p2);
void n2com(struct sockaddr_in address, int *a1, int *a2, int *a3, int *a4,
	   int *p1, int *p2);
void closeandexit(int status);
struct sockaddr_in extract_address(sstr * buf);
void write_log(int priority, char *msg, ...);
sstr *addr2name(struct in_addr address);
void quit(exitreason reason);
int append_read(int fd, char *buf);
void sstrerr(void);

/*****************
**linux.c/bsd.c **
******************/
int get_orig_dest(int fd, struct sockaddr_in *addr);
int get_local_address(struct sockaddr_in *addr);
int bindtodevice(int fd);
#ifdef TRANS_DATA
int transdata_setup(void);
int transdata_needprivs(void);
int transp_connect(struct sockaddr_in dest, struct sockaddr_in src);
int intercept_listen(struct sockaddr_in intercept,
		     struct sockaddr_in listen_on, socketuse use);
int il_free(void);
#else
static inline int transdata_setup(void)
{
	return -1;
};
static inline int transdata_needprivs(void)
{
	return FALSE;
};
static inline int transp_connect(struct sockaddr_in dest,
				 struct sockaddr_in src)
{
	return -1;
};
static inline int intercept_listen(struct sockaddr_in intercept,
				   struct sockaddr_in listen_on,
				   socketuse use)
{
	return -1;
};
static inline int il_free(void)
{
	return -1;
};
#endif

/*****************
**     ntp.c    **
******************/
#ifdef NON_TRANS_PROXY
void ntp_changedest(void);
void ntp_senduser(void);
#else
static inline void ntp_changedest(void)
{
};
static inline void ntp_senduser(void)
{
};
#endif

/*****************
**    cache.c   **
******************/
#ifdef USE_CACHE
int cache_geninit(void);
void cache_init(void);
int cache_parsed_reply(int code, sstr * msg);
void cache_inc_data(sstr * inc);
int cache_close_data(void);
#else
static inline int cache_geninit(void)
{
	return (0);
};
static inline void cache_init(void)
{
};
static inline int cache_parsed_reply(int code, sstr * msg)
{
	return FALSE;
};
static inline void cache_inc_data(sstr * inc)
{
};
static inline int cache_close_data(void)
{
	return -1;
};
#endif

/*******************
**Global Variables**
********************/
session_info *info;

#endif				/* COMMON_H */
