/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
ftp-proxy.c -- central code for managing forwarding.
               Also the data path forwarding stuff.
  ***************************************/


#include <sys/ioctl.h>
#include <fcntl.h>
#include <syslog.h>

#include "ftp-proxy.h"
#include "control.h"
#include "ftp-cmds.h"

void run_proxy(void);

void client_data_connect(void);
void server_data_connect(void);

void writebuf2client(void);
void writebuf2server(void);
void forwarddata2client(void);
void forwarddata2server(void);

void closecd(void);
void check_closesd(void);

/* ------------------------------------------------------------- **
**  Sets up the control connection to the server and initialises
**  session wide parts of the info structure.
** ------------------------------------------------------------- */
void init_session(int fd, struct sockaddr_in source)
{
	sstr *tmp;

	info = (session_info *) malloc(sizeof(session_info));
	if (info == NULL) quit(MALLOC);
	memset(info, 0, sizeof(session_info));

	info->listen = info->client_data.fd = info->server_data.fd = -1;
	info->server_control.fd = -1;

	info->server_control.buf = sstr_init(BUF_LEN);
	info->client_control.buf = sstr_init(BUF_LEN);
	info->server_data.buf = sstr_init(BUF_LEN);
	info->client_data.buf = sstr_init(BUF_LEN);

	if (config.apconv) info->mode = APCONV;
	else info->mode = ACTIVE;
	info->state = NEITHER;

	info->client_control.address = source;
	info->client_control.fd = fd;
	get_orig_dest(fd, &info->server_control.address);

	ntp_changedest();

	/*Check for loops. Won't work if Listen undefined*/
	if(info->server_control.address.sin_addr.s_addr
	   == config.listen_address.sin_addr.s_addr &&
	   info->server_control.address.sin_port
	   == config.listen_address.sin_port) {
		write_log(ERROR, "Attempt to connect to self");
		send_cmessage(421, "Proxy tried to loop. Closing connection");
		exit(0);
	}

	tmp = sstr_init(MAX_LINE_LEN);
	sstr_cpy(tmp, addr2name(info->client_control.address.sin_addr));
	sstr_ncat2(tmp, " to ", 4);
	sstr_cat(tmp, addr2name(info->server_control.address.sin_addr));
	write_log(INFO, "Connect from %s", sstr_buf(tmp));
	sstr_free(tmp);

	if (!config_connectionok(&info->client_control.address,
				 &info->server_control.address)) {
		write_log(ATTACK, "Denied by ACLs.");
		send_cmessage(501, "Connection denied. Bye");
		close(info->client_control.fd);
		exit(0);
	}

	debug("Connecting to server...");

	if ((info->server_control.fd =
	     connect_to_socket(info->server_control.address, CTRL)) == -1) {
		write(info->client_control.fd,
		      "501 Proxy unable to contact ftp server\n", 39);
		write_log(ERROR,
			  "Connection closed -- unable to contact server");
		close(info->client_control.fd);
		exit(0);
	}

	debug("OK\n");

	ftpcmds_init();		/* Needs doing before cache_init() */
	if (config.cachemod)
		cache_init();	/*Needs doing before ntp_senduser() */

	ntp_senduser();

	run_proxy();
}

/* ------------------------------------------------------------- **
**  The main loop for each session.
** ------------------------------------------------------------- */
void run_proxy()
{
	int i;

	do {
		i = get_control_line(GET_BOTH);
		if (i & GET_CLNT) client_control_forward();
		if (i & GET_SRVR) server_control_forward();
	} while (TRUE);
}

/* ------------------------------------------------------------- **
** Setup fd sets for reading/writing. We always listen to the client's
** control stream, and other fds are selected on depending on state. 
** ------------------------------------------------------------- */
int setup_fds(fd_set * reads, fd_set * writes)
{
	int n;

	FD_ZERO(reads);
	FD_ZERO(writes);

	FD_SET(info->client_control.fd, reads);
	n=info->client_control.fd;

	if (info->server_control.fd != -1) {
		FD_SET(info->server_control.fd, reads);
		if(info->server_control.fd>n) n=info->server_control.fd;
	}

	if (info->server_data.fd != -1) {
		if (sstr_len(info->client_data.buf) == 0
		    && info->state == DOWNLOAD)
			FD_SET(info->server_data.fd, reads);

		if (sstr_len(info->server_data.buf) != 0)
			FD_SET(info->server_data.fd, writes);
		if(info->server_data.fd>n) n=info->server_data.fd;
	}
	if (info->client_data.fd != -1) {
		if (sstr_len(info->server_data.buf) == 0
		    && info->state == UPLOAD)
			FD_SET(info->client_data.fd, reads);
		if (sstr_len(info->client_data.buf) != 0)
			FD_SET(info->client_data.fd, writes);
		if(info->client_data.fd>n) n=info->client_data.fd;
	}
	if (info->listen != -1) {
		FD_SET(info->listen, reads);
		if(info->listen > n) n=info->listen;
	}
	return(n);
}

/* ------------------------------------------------------------- **
** Check the fd_sets, and do data connection forwarding if any.
** ------------------------------------------------------------- */
void do_dataforward(fd_set * reads, fd_set * writes)
{
	if (info->listen != -1 && FD_ISSET(info->listen, reads)) {
		if (info->mode == ACTIVE) server_data_connect();
		else client_data_connect();
	}

	if (info->server_data.fd != -1)
		if (FD_ISSET(info->server_data.fd, reads))
			forwarddata2client();
	if (info->server_data.fd != -1)
		if (FD_ISSET(info->server_data.fd, writes))
			writebuf2server();

	if (info->client_data.fd != -1)
		if (FD_ISSET(info->client_data.fd, reads))
			forwarddata2server();
	if (info->client_data.fd != -1)
		if (FD_ISSET(info->client_data.fd, writes))
			writebuf2client();

}

/* ------------------------------------------------------------- **
**  Client just connected to the data line
** ------------------------------------------------------------- */
void client_data_connect()
{
	int len = sizeof(info->client_data.address);

	debug("  Client has connected to proxy data line\n");
	info->client_data.fd = accept(info->listen,
				      (struct sockaddr *) &info->
				      client_data.address, &len);

	if (config.sameaddress) {
		if (info->client_data.address.sin_addr.s_addr !=
		    info->client_control.address.sin_addr.s_addr) {
			write_log(ATTACK,
				  "Blocked %s from connecting to data line",
				  addr2name(info->client_data.address.
					    sin_addr));
			close(info->client_data.fd);
			return;
		}
	}
	il_free();		/* Remove the ipchains entry for intercepting this
				 * data connection.*/

	/* Connect to server (if cache code hasn't already done so) */
	if (info->server_data.fd == -1) {
		info->server_data.fd =
		    connect_to_socket(info->server_data.address, PASV);
		if (info->server_data.fd == -1) {
			write_log(ERROR, "Failed to contact server data port");
			close(info->client_data.fd);
			info->client_data.fd = -1;
			return;
		}
		sstr_empty(info->server_data.buf);
	}
	sstr_empty(info->client_data.buf);
}

/* ------------------------------------------------------------- **
**  Server just connected to the data line
** ------------------------------------------------------------- */
void server_data_connect()
{
	int len = sizeof(info->server_data.address);

	debug("  Server has connected to proxy data line\n");
	info->server_data.fd = accept(info->listen,
				      (struct sockaddr *) &info->
				      server_data.address, &len);
	if (config.sameaddress) {
		if (info->server_data.address.sin_addr.s_addr !=
		    info->server_control.address.sin_addr.s_addr) {
			write_log(ATTACK,
				  "Blocked %s from connecting to data line",
				  addr2name(info->server_data.address.
					    sin_addr));
			close(info->server_data.fd);
			info->server_data.fd = -1;
			return;
		}
	}

	if (config.transdata)
		info->client_data.fd =
		    transp_connect(info->client_data.address,
				   info->server_data.address);
	else
		info->client_data.fd =
		    connect_to_socket(info->client_data.address, ACTV);
	if (info->client_data.fd == -1) {
		write_log(ERROR, "Failed to contact client data port");
		close(info->server_data.fd);
		info->server_data.fd = -1;
		return;
	}

	sstr_empty(info->server_data.buf);
	sstr_empty(info->client_data.buf);
}

/* ------------------------------------------------------------- **
** Forward as much data as possible from client-->server, and store
** the rest in server_data.buf. 
** ------------------------------------------------------------- */
void forwarddata2server()
{
	int i, j;

	i = sstr_append_read(info->client_data.fd, info->server_data.buf, 0);
	if (i < 1) {
		closecd();
		return;
	}
	j = sstr_write(info->server_data.fd, info->server_data.buf, 0);
	if (j == -1) {
		debug_perr("write");
		if (errno == EPIPE) {
			check_closesd();
			return;
		}
		j = 0;
	}
	sstr_split(info->server_data.buf, NULL, 0, j);
	debug(".");
}

/* ------------------------------------------------------------- **
** Forward as much data as possible from server-->client, and store
** the rest in client_data.buf. 
** ------------------------------------------------------------- */
void forwarddata2client()
{
	int j;

	sstr_append_read(info->server_data.fd, info->client_data.buf, 0);

	if (sstr_len(info->client_data.buf) < 1) {
		check_closesd();
		return;
	}

	cache_inc_data(info->client_data.buf);

	j = sstr_write(info->client_data.fd, info->client_data.buf, 0);
	if (j == -1) {
		 /*FIXME*/ debug_perr("write");
		if (errno == EPIPE) {
			closecd();
			return;
		}
		j = 0;
	}
	sstr_split(info->client_data.buf, NULL, 0, j);
	debug(".");
}

/* ------------------------------------------------------------- **
**  Flush as much buffer as we can
** ------------------------------------------------------------- */
void writebuf2client()
{
	int i;

	i = sstr_write(info->client_data.fd, info->client_data.buf, 0);
	if (i == -1 && errno == EPIPE) {
		closecd();
		return;
	}
	sstr_split(info->client_data.buf, NULL, 0, i);

	if (sstr_len(info->client_data.buf) == 0 && info->server_data.fd == -1)
		closecd();
	debug("_");
}

/* ------------------------------------------------------------- **
**  Flush as much buffer as we can
** ------------------------------------------------------------- */
void writebuf2server()
{
	int i;

	i = sstr_write(info->server_data.fd, info->server_data.buf, 0);
	if (i == -1 && errno == EPIPE) {
		check_closesd();
		return;
	}
	sstr_split(info->server_data.buf, NULL, 0, i);

	if (sstr_len(info->server_data.buf) == 0 && info->client_data.fd == -1)
		check_closesd();
	debug("_");
}

/* ------------------------------------------------------------- **
**  Close data socket, and if there is nothing left to flush close
** the other one too.
** ------------------------------------------------------------- */
void closecd(void)
{
	debug("Closing client data connection\n");
	info->state = NEITHER;
	close(info->client_data.fd);
	info->client_data.fd = -1;
	if (sstr_len(info->server_data.buf) == 0) {
		debug("Closing server data connection\n");
		close(info->server_data.fd);
		info->server_data.fd = -1;
		cache_close_data();
	}
}

/* ------------------------------------------------------------- **
** Renamed from closesd() since we don't necessarily close the server
** data connection. Caching code is allowed to return data in several
** files, in which case cace_close_data() will return the next fd. It
** returns (-1) when done.
** ------------------------------------------------------------- */
void check_closesd(void)
{
	int tmp;
	debug("Closing server data connection\n");

	close(info->server_data.fd);

	tmp = cache_close_data();
	if (tmp != -1) {
		tmp = dup2(tmp, info->server_data.fd);
		if (tmp == -1)
			fprintf(stderr, "arrghh\n");
		return;
	}

	info->state = NEITHER;
	info->server_data.fd = -1;
	if (sstr_len(info->client_data.buf) == 0) {
		debug("Closing client data connection\n");
		close(info->client_data.fd);
		info->client_data.fd = -1;
	}
}
